/*
 * Copyright (c) 1999, 2000 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <netdb.h>
#include <syslog.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <unistd.h>

#include <getarg.h>
#include <err.h>

#include <roken.h>

#ifndef RCSID
#define RCSID(msg) \
static /**/const char *const rcsid[] = { (char *)rcsid, "\100(#)" msg }
#endif

RCSID("$Id: krb-forward.c,v 1.9 2000/03/31 16:56:32 lha Exp $");

#define MAX_SERVERS	10
#define MAX_PACKET_SIZE 1500
#define NSOCKETS	10
#define KRB_PORT	750
#define MAX_ERRORS	10

struct client {
    int fd;
    struct sockaddr_in addr;
    int inusep;
};

/*
 * A server
 */

struct krb_server {
    char *name;
} servers[MAX_SERVERS];

static int nserver = 0;
static int current_server = -1;
static int server_changes = 0;
static int server_error = 0;
static struct sockaddr_in to_addr;

/*
 *
 */

unsigned int kerberos_port = KRB_PORT;

unsigned int krb_server = 0;
unsigned int debug = 0;
unsigned int version = 0;
unsigned int help = 0;
unsigned int arg_fork = 1;

int nsocket = NSOCKETS;		/* number of clients to queue from */
struct client *clients;		/* the clients */


#define XDEBUG(x) do { if (debug) printf x ; } while (0) 

struct getargs args[] = {
    { "kerberos-port", 'p', arg_integer, &kerberos_port, "port to use", NULL },
    { "clients", 'c', arg_integer, &nsocket, "number of clients to use", NULL },
    { "version", 'v', arg_flag, &version, "version", NULL },
    { "help", 'h', arg_flag, &help, "version", NULL },
    { "fork", 'f', arg_negative_flag, &arg_fork, "don't fork", NULL},
    { "debug", 'd', arg_flag, &debug, "print debugging output", NULL }
};

/*
 * Setup all client sockets
 */

static void
setup_sockets (void)
{
    int i;

    clients = malloc (nsocket * sizeof(struct client));
    if (clients == NULL)
	err (1, "setup_sockets: malloc");

    for (i = 0; i < nsocket; i++)
	clients[i].inusep = 0;
}

/*
 * Change to a new server in the ``servers'' list
 */

static void
setup_to_addr (void)
{
    struct hostent *he;
    
    server_changes++;
    server_error = 0;
    current_server++;
    if (current_server == nserver)
	current_server = 0;
    
    syslog (LOG_ERR, 
	    "changed for the %d time kerberos server (this time to %s)",
	    server_changes,
	    servers[current_server].name);
	
    he = gethostbyname (servers[current_server].name);
    if (he == NULL)
	errx (1, "no such server %s", servers[current_server].name);
    

    to_addr.sin_family = AF_INET;
    to_addr.sin_addr.s_addr = *(unsigned long *) he->h_addr_list[0];
    to_addr.sin_port = htons(kerberos_port);
}

/*
 * Bind to the kerberos port and return the fd to the socket.
 */

static int
setup_krb_socket (void)
{
    struct sockaddr_in addr;
    int fd, ret;

    fd = socket (AF_INET, SOCK_DGRAM, 0);
    if (fd < 0)
	err (1, "setup_krb_socket: socket");

    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = INADDR_ANY;
    addr.sin_port = htons(kerberos_port);

    ret = bind (fd, (struct sockaddr *) &addr, sizeof (addr));
    if (ret)
	err (1, "setup_krb_socket: bind");

    return fd;
}

/*
 * delete an existing active client
 */

static void
close_client (struct client *c)
{
    close (c->fd);
    c->inusep = 0;
}

/*
 * start using this client
 */

static void
start_client (struct client *c)
{
    int ret;
    struct sockaddr_in addr;
    
    c->inusep = 1;
    c->fd = socket (AF_INET, SOCK_DGRAM, 0);
    if (c->fd < 0)
	err (1, "socket AF_INET SOCK_DGRAM");
    memset (&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    ret = bind (c->fd, (struct sockaddr *)&addr, sizeof(addr));
    if (ret)
	err (1, "bind");
}

/*
 * Find a new client to store where the request came from.
 *
 * If we have to reuse a old already in use struct consider the old server
 * a slow/down server and change to a new server.
 */

static struct client *
find_new_client (void)
{
    int i;
    struct client *ret;

    for (i = 0; i < nsocket; i++) {
	if (!clients[i].inusep) {
	    ret = &clients[i];
	    start_client (ret);
	    return ret;
	}
    }

    /*
     * If error/slowserver to large, change server
     */

    if (server_error++ > MAX_ERRORS)
	setup_to_addr();
    ret = &clients[rand() % nsocket];
    close_client (ret);
    start_client (ret);
    return ret;
}

/*
 * Main program, parse arguments and drive the main loop
 */

int 
main (int argc, char **argv)
{
    int krb_fd, max_fd = -1, ret, i;
    char packet[MAX_PACKET_SIZE];
    fd_set fds;
    unsigned long numpacket = 0;
    int optind = 0;
    
    set_progname (argv[0]);
    openlog (argv[0], LOG_PID, LOG_DAEMON);
    srand (time(NULL));

    /*
     * Parse argument
     */

    if (getarg (args, sizeof(args)/sizeof(*args), argc, argv, &optind)) {
	arg_printusage (args, sizeof(args)/sizeof(*args), 
			argv[0], "kerberos-server [kerberos-server2 ...]");
	errx (1, "bad argument: %s", argv[optind]);
    }

    if (version)
	errx (0, "version: "
	      "$Id: krb-forward.c,v 1.9 2000/03/31 16:56:32 lha Exp $");

    if (help) {
	arg_printusage (args, sizeof(args)/sizeof(*args), 
			argv[0], "nothing here");
	exit (0);
    }	

    argv += optind;
    argc -= optind;

    /*
     * Setup servers
     */

    while (argc) {
	if (nserver == MAX_SERVERS)
	    errx (1, "too many servers");

	servers[nserver].name = strdup(*argv);
	if (servers[nserver].name == NULL)
	    err (1, "strdup");
	nserver++;
	argc--;
	argv++;
    }

    if (nserver == 0)
	errx (1, "No server given");

    setup_sockets ();
    setup_to_addr();
    krb_fd = setup_krb_socket ();

    if (arg_fork)
	daemon(0, 0);

    XDEBUG(("staring service\n"));

    /*
     * Main loop
     */

    while (1) {
    
	FD_ZERO(&fds);
	for (i = 0; i < nsocket; i++) {
	    if (clients[i].inusep) {
		FD_SET(clients[i].fd, &fds);
		if (clients[i].fd > max_fd)
		    max_fd = clients[i].fd;
	    }
	}
	FD_SET(krb_fd, &fds);
	if (krb_fd > max_fd)
	    max_fd = krb_fd;

	ret = select (max_fd + 1, &fds, NULL, NULL, NULL);
	if (ret < 0) {
	    if (errno == EINTR)
		continue;
	    else
		err (1, "select");
	}

	XDEBUG(("got a packet\n"));

	/*
	 * We got a answer from the server, send to the asking client
	 */

	for (i = 0; i < nsocket; i++) {
	    if (clients[i].inusep && FD_ISSET(clients[i].fd, &fds)) {
		XDEBUG(("got answer from the krb-server to client #%d/%s:%d\n",
			i, inet_ntoa(clients[i].addr.sin_addr),
			ntohs(clients[i].addr.sin_port)));

		ret = recvfrom (clients[i].fd, &packet, sizeof(packet),
				0, NULL, 0);
		if (ret <= 0) {
		    close_client (&clients[i]);
		    syslog (LOG_ERR, "recvfrom on client %d: %s",
			    i, strerror(errno));
		    continue;
		}
		ret = sendto (krb_fd, &packet, ret, 0,
			      (struct sockaddr *)&clients[i].addr,
			      sizeof(clients[i].addr));
		close_client (&clients[i]);
		if (ret < 0) {
		    syslog (LOG_ERR, "sendto on client %s %d: %s",
			    inet_ntoa(clients[i].addr.sin_addr),
			    i, strerror(errno));
		    continue;
		}
	    }
	}

	/*
	 * We got a request from a client, send to the krb-server
	 */

	if (FD_ISSET(krb_fd, &fds)) {
	    struct client *c = find_new_client();
	    int size = sizeof(c->addr);
	    
	    ret = recvfrom (krb_fd, &packet, sizeof(packet),
			    0, (struct sockaddr *)&c->addr, &size);
	    if (ret < 0) {
		close_client (c);
		syslog (LOG_ERR, "recvfrom on krb_fd: %s",
			strerror(errno));
		continue;
	    }
	    
	    XDEBUG(("got request from the client to client %s:%d\n",
		    inet_ntoa(c->addr.sin_addr),
		    ntohs(c->addr.sin_port)));
		   
	    ret = sendto (c->fd, &packet, ret, 0, 
			  (struct sockaddr *) &to_addr, sizeof (to_addr));
	    if (ret < 0) {
		close_client (c);
		syslog (LOG_ERR, "sendto on krb_fd %s: %s",
			inet_ntoa(to_addr.sin_addr), strerror(errno));
		continue;
	    }
	    XDEBUG(("sending packet to the krb-server\n"));
	    numpacket++;
	}
    }
    return 0;
}
